open Coqlib
open Declare
open Decl_kinds
open Entries
open Impargs
open Libnames
open Names
open Pp
open Rawterm
open Tacexpr
open Term
open Topconstr
open Util
(*open Indfun_common*)

let ffcpo_msg = "ffcpo"

let ffcpo_modules = [
  ["Coq"; "ffcpo"; "Preorder"];
  ["Coq"; "ffcpo"; "CompletePreorder"];
  ["Coq"; "ffcpo"; "FlatCompletePreorder"];
  ["Coq"; "ffcpo"; "FixedPoint"];
  ["Coq"; "ffcpo"; "Minimisation"]]

let coq_constant dir s =
  (* (Coqlib.gen_constant_in_modules ffcpo_msg *)
  Coqlib.gen_constant ffcpo_msg dir s
(*    (Coqlib.init_modules @ ffcpo_modules) s) *)

let ffcpo_constant dir s =
  Coqlib.gen_constant ffcpo_msg ("ffcpo"::dir) s

let expr_of_constr c =
  Constrextern.extern_constr false (Global.env()) c

(* ffcpo library constants *)
let coq_ord = lazy (ffcpo_constant ["Preorder"] "ord")
let coq_carrier_ord = lazy (ffcpo_constant ["Preorder"] "carrier_ord")
let coq_funord = lazy (ffcpo_constant ["Preorder"] "funord")
let coq_monotonic = lazy (ffcpo_constant ["Preorder"] "monotonic")
let coq_f_mono = lazy (ffcpo_constant ["Preorder"] "f_mono")
let coq_Build_monofun = lazy (ffcpo_constant ["Preorder"] "Build_monofun")
let coq_monofunord = lazy (ffcpo_constant ["Preorder"] "monofunord")
let coq_cpo = lazy (ffcpo_constant ["CompletePreorder"] "cpo")
let coq_carrier_cpo = lazy (ffcpo_constant ["CompletePreorder"] "O_cpo")
let coq_continuous = lazy (ffcpo_constant ["CompletePreorder"] "continuous")
let coq_Build_contfun = lazy (ffcpo_constant ["CompletePreorder"] "Build_contfun")
let coq_funcpo = lazy (ffcpo_constant ["CompletePreorder"] "funcpo")
let coq_contfuncpo = lazy (ffcpo_constant ["CompletePreorder"] "contfuncpo")
let coq_contfun_monofun = lazy (ffcpo_constant ["CompletePreorder"] "contfun_monofun")
let coq_partial = lazy (ffcpo_constant ["FlatCompletePreorder"] "partial")
let coq_flat_ord = lazy (ffcpo_constant ["FlatCompletePreorder"] "flat_ord")
let coq_flat_cpo = lazy (ffcpo_constant ["FlatCompletePreorder"] "flat_cpo")
let coq_fixp = lazy (ffcpo_constant ["FixedPoint"] "fixp")

(* Coq's constants *)
let coq_proj1 = lazy (coq_constant ["Init"; "Logic"] "proj1")
let coq_proj2 = lazy (coq_constant ["Init"; "Logic"] "proj2")

let mkProj1 a b c = mkApp (Lazy.force coq_proj1, [| a; b; c |])
let mkProj2 a b c = mkApp (Lazy.force coq_proj2, [| a; b; c |])

let envn : Environ.env ref =
  ref Environ.empty_env
let c : constr ref =
  ref (mkRel 1)
let e : constr_expr ref =
  ref (CHole (Util.dummy_loc, None))
let r : rawconstr ref =
  ref (RHole (Util.dummy_loc, Evd.InternalHole))
let i : (Topconstr.explicitation * (bool * bool)) list ref = (*{contents = []}*)
  ref [ExplByName(id_of_string "dummy"), (false, false)]

(* lift a plain type to a preorder on functions *)
let lift_ord_type (binders:local_binder list) (typ:constr_expr) =
  let env = Global.env() in
  let coq_partial = Constrextern.extern_constr false env (Lazy.force coq_partial) in
  let coq_funord = Constrextern.extern_constr false env (Lazy.force coq_funord) in
  let coq_flat_ord = Constrextern.extern_constr false env (Lazy.force coq_flat_ord) in
  let typ = Command.generalize_constr_expr typ binders in
  let loc = Util.dummy_loc in
  let rec cvt_type_inside
    (bl:(name located list * binder_kind * constr_expr) list)
    (typB:constr_expr)
    : constr_expr =
  match bl with
    (names, _, typA)::tl ->
      let folder _ b =
        CApp(loc, (None, coq_funord), [typA, None; b, None])
      in
        (List.fold_right folder names (cvt_type_inside tl typB))
  | [] -> cvt_type typB
  and cvt_type : constr_expr -> constr_expr = function
    CProdN(_, bl, typ) -> cvt_type_inside bl typ
  | CArrow(_, typA, typB) ->
      CApp(loc, (None, coq_funord), [typA, None; cvt_type typB, None])
  | CApp(lc, (o, coq_partial), [c, e]) -> CApp(lc, (o, coq_flat_ord), [c, e])
(*  | typB -> CApp(loc, (None, coq_flat_ord), [typB, None]) *)
  | _ -> raise (Util.UserError("lift_ord_type", str "lift_ord_type: unsupported typing"))
  in
  let result = cvt_type typ in
    result

(* lift a plain type to a _complete_ preorder on functions; can be merged with lift_ord_type *)
let lift_cpo_type (binders:local_binder list) (typ:constr_expr) =
  let env = Global.env() in
  let coq_partial = Constrextern.extern_constr false env (Lazy.force coq_partial) in
  let coq_funcpo = Constrextern.extern_constr false env (Lazy.force coq_funcpo) in
  let coq_flat_cpo = Constrextern.extern_constr false env (Lazy.force coq_flat_cpo) in
  let typ = Command.generalize_constr_expr typ binders in
  let loc = Util.dummy_loc in
  let rec cvt_type_inside
    (bl:(name located list * binder_kind * constr_expr) list)
    (typB:constr_expr)
    : constr_expr =
  match bl with
    (names, _, typA)::tl ->
      let folder _ b =
        CApp(loc, (None, coq_funcpo), [typA, None; b, None])
      in
        (List.fold_right folder names (cvt_type_inside tl typB))
  | [] -> cvt_type typB
  and cvt_type : constr_expr -> constr_expr = function
    CProdN(_, bl, typ) -> cvt_type_inside bl typ
  | CArrow(_, typA, typB) ->
      CApp(loc, (None, coq_funcpo), [typA, None; cvt_type typB, None])
  | CApp(lc, (o, coq_partial), [c, e]) -> CApp(lc, (o, coq_flat_cpo), [c, e])
(*  | typB -> CApp(loc, (None, coq_flat_cpo), [typB, None]) *)
  | _ -> raise (Util.UserError("lift_cpo_type", str "lift_cpo_type: unsupported typing"))
  in
  let result = cvt_type typ in
    result


let rec abstract_rawconstr c = function
  | [] -> c
  | Topconstr.LocalRawDef (x,b)::bl -> Topconstr.mkLetInC(x,b,abstract_rawconstr c bl)
  | Topconstr.LocalRawAssum (idl,k,t)::bl ->
      List.fold_right (fun x b -> Topconstr.mkLambdaC([x],k,t,b)) idl
        (abstract_rawconstr c bl)

let interp_casted_constr_with_implicits sigma env imps c =
(*  Constrintern.interp_rawconstr_with_implicits sigma env [] impls c *)
  Constrintern.intern_gen false sigma env ~impls:([],imps) 
    ~allow_patvar:false  ~ltacvars:([],[]) c

(* provides a name for the functional *)
(* FIXME: check for freshness *)
let f_id_of_id id =
  id_of_string ("f_"^(string_of_id id))

(* this function's values are not used (yet?) *)
(* build a Rawterm object and a list of implicits given an ffcpo_expr *)
let build_ffcpo ((id:identifier), (binders:local_binder list), (typ:constr_expr), (body:constr_expr)) =
  let env_glob = Global.env() in
  (* let body = subst_in_constr_expr id f_id body in *)
  let (_sign, _imps) = (env_glob, [])
(*
    (fun (env, imps) (* (id, binders, type_fun, _) *) ->
      let typ = Command.generalize_constr_expr typ binders in
      let arity = Constrintern.interp_type Evd.empty env_glob typ in
      let imp =
        if Impargs.is_implicit_args()
        then Impargs.compute_implicits env_glob arity
        else [] in
      let imps' = (id, ([], imp, Notation.compute_arguments_scope arity))::imps in
        (Environ.push_named (id, None, arity) (env, imps')))
    (env_glob, []) (* ffcpo_def *)
*)
  in
  let _def =
    (* Declare local notations *)
    let fs = States.freeze() in
    let def = try
    let def = abstract_rawconstr body binders in
      interp_casted_constr_with_implicits Evd.empty _sign _imps def
	with e -> States.unfreeze fs; raise e
    in States.unfreeze fs; def
  in _def, _imps

(* declare a functional; [binders] should contain exactly one argument of type [typ] *)
(* TODO: restrict the code to handle only the expected [binders] *)
let declare_functional ((id:identifier), (binders:local_binder list), (typ:constr_expr), (body:constr_expr)) (*def*) imps =
  (* let evdref = ref (Evd.create_evar_defs Evd.empty) in *)
  (* let type_of_f = Command.generalize_constr_expr f.ffcpo_type f.ffcpo_binders in *)
  let f_id = f_id_of_id id in
  let sigma = Evd.empty in
  let env = Global.env() in
    (* TODO: move printing the definition after [declare_constant] *)
    Pp.msg ((Ppconstr.pr_id f_id) ++ pr_str " = ");
  let bodyc = Command.abstract_constr_expr body binders in
    Pp.msgnl (Ppconstr.pr_constr_expr bodyc);
  let arityc = Command.generalize_constr_expr typ binders in
    Pp.msgnl (pr_str "  : " ++ Ppconstr.pr_constr_expr arityc);
  let ce = {
    const_entry_body = Constrintern.interp_constr sigma env bodyc;
    const_entry_type = Some (Constrintern.interp_type sigma env arityc);
    const_entry_opaque = false;
    const_entry_boxed = Flags.boxed_definitions()
  } in
  let kn = declare_constant f_id (DefinitionEntry ce, Decl_kinds.IsDefinition Decl_kinds.Definition) in
  let gr = Libnames.ConstRef kn in
    if Impargs.is_implicit_args () || imps <> [] then
      declare_manual_implicits false gr (Impargs.is_implicit_args ()) imps;
    gr

(* put carrier_ord on top of a constr expression *)
let coerce_ord (c:constr) =
  mkApp (Lazy.force coq_carrier_ord, Array.of_list [c])

(* put O_cpo on top of a constr expression *)
let coerce_cpo (c:constr) =
  mkApp (Lazy.force coq_carrier_cpo, Array.of_list [c])

let declare_constant_hooked (id:identifier) (typ:constr) (body:constr) (hook:declaration_hook) =
  let k = declare_constant
            id
            ( DefinitionEntry
              { const_entry_body = body;
                const_entry_type = Some typ;
                const_entry_opaque = false;
                const_entry_boxed = Flags.boxed_definitions() },
              IsDefinition Definition )
  in hook Global (ConstRef k)

(* ATTENTION: this is a commented function; see the function declare_ffcpo below
(* declares a recursive function as the least fixed point of its continuous functional *)
(* NOTE: This function asks a Coq user to prove two lemmas subsequently, with the second
   lemma appearing after the first have been saved. This looks a bit clumsy from Coq. *)
let declare_ffcpo_retro (id:identifier) (f:global_reference) (typ:constr_expr) =
  let f = constr_of_global f in
  let sigma = Evd.empty in
  let env = Global.env () in
  let ord_typ = Constrintern.interp_constr sigma env (lift_ord_type [] typ) in
  let cpo_typ = Constrintern.interp_constr sigma env (lift_cpo_type [] typ) in
  let string_of_f = string_of_id id in
  let id_monotonic = id_of_string (string_of_f ^ "_monotonic") in
  let lemma_monotonic = mkApp (Lazy.force coq_monotonic, [|ord_typ; ord_typ; f|]) in
  let goal_kind = Decl_kinds.Global, Decl_kinds.DefinitionBody Decl_kinds.Definition in
  let prove_monotonic = Command.start_proof id_monotonic goal_kind lemma_monotonic in
  let f_monotonic mono_proof =
    mkApp (Lazy.force coq_Build_monofun, [|ord_typ; ord_typ; f; constr_of_global mono_proof|]) in
  let f_monotonic_typ =
    coerce_ord (mkApp (Lazy.force coq_monofunord, [|ord_typ; ord_typ|])) in
  let id_mono_f = id_of_string ("mono_" ^ string_of_f) in
  let declare_monotonic (mono_proof:global_reference) (hook:declaration_hook) =
    declare_constant_hooked id_mono_f f_monotonic_typ (f_monotonic mono_proof) hook
  in
  let id_continuous = id_of_string (string_of_f ^ "_continuous") in
  let lemma_continuous mono_f =
    mkApp (Lazy.force coq_continuous, [|cpo_typ; cpo_typ; constr_of_global mono_f|]) in
  let prove_continuous mono_f = Command.start_proof id_continuous goal_kind (lemma_continuous mono_f) in
  let f_continuous mono_f cont_proof = mkApp (Lazy.force coq_Build_contfun,
    [|cpo_typ; cpo_typ; constr_of_global mono_f; constr_of_global cont_proof|]) in
  let f_continuous_typ =
    coerce_ord (coerce_cpo (mkApp (Lazy.force coq_contfuncpo, [|cpo_typ; cpo_typ|]))) in
  let id_cont_f = id_of_string ("cont_" ^ string_of_f) in
  let declare_continuous (mono_f:global_reference) (cont_proof:global_reference) (hook:declaration_hook) =
    declare_constant_hooked id_cont_f f_continuous_typ (f_continuous mono_f cont_proof) hook
  in
  let f_lfp cont_f =
    mkApp (Lazy.force coq_f_mono,
     [|
       coerce_cpo (mkApp (Lazy.force coq_contfuncpo, [|cpo_typ; cpo_typ|]));
       coerce_cpo cpo_typ;
       mkApp (Lazy.force coq_contfun_monofun,
         [|
           (mkApp (Lazy.force coq_contfuncpo, [|cpo_typ; cpo_typ|]));
           cpo_typ;
           mkApp (Lazy.force coq_fixp, [|cpo_typ|])
         |]);
       constr_of_global cont_f
     |]) in
  let declare_lfp (cont_f:global_reference) (hook:declaration_hook) =
    declare_constant_hooked id (coerce_ord (coerce_cpo cpo_typ)) (f_lfp cont_f) hook
  in
  let hook_done _ _ = () in
  let hook_lfp _ cont_f = (declare_lfp cont_f hook_done; ()) in
  let hook_declare_continuous mono_f _ cont_proof = (declare_continuous mono_f cont_proof hook_lfp; ()) in
  let hook_continuous _ mono_f = ((prove_continuous mono_f) (hook_declare_continuous mono_f); ()) in
  let hook_declare_monotonic _ mono_proof = (declare_monotonic mono_proof hook_continuous; ()) in
    prove_monotonic hook_declare_monotonic;
    Flags.if_verbose (fun () -> msg (Printer.pr_open_subgoals ())) ()
    (* if !pcoq <> None then (Option.get !pcoq).start_proof () *)
*)

(* declares a recursive function as the least fixed point of its continuous functional *)
(* this function asks to prove a single conjunctive goal rather than two subsequent ones *)
let declare_ffcpo (id:identifier) (f:global_reference) (typ:constr_expr) =
  let f = constr_of_global f in
  let sigma = Evd.empty in
  let env = Global.env () in
  let ord_typ = Constrintern.interp_constr sigma env (lift_ord_type [] typ) in
  let cpo_typ = Constrintern.interp_constr sigma env (lift_cpo_type [] typ) in
  let string_of_f = string_of_id id in
  let lemma_monotonic = mkApp (Lazy.force coq_monotonic, [|ord_typ; ord_typ; f|]) in
  let goal_kind = Decl_kinds.Global, Decl_kinds.DefinitionBody Decl_kinds.Definition in
  let id_mono_proof = id_of_string "mono_proof" in
  let lemma_continuous =
    mkProd (Name id_mono_proof, lemma_monotonic, (mkApp (Lazy.force coq_continuous,
            [|
              cpo_typ;
              cpo_typ;
              mkApp (Lazy.force coq_Build_monofun,
              [|
                ord_typ;
                ord_typ;
                f;
                mkRel 1 (* id_mono_proof *)
              |])
            |])))
  in
  let lemma_monotonic_continuous = mkApp (build_coq_and(), [|lemma_monotonic; lemma_continuous|]) in
  let id_monotonic_continuous = id_of_string (string_of_f ^ "_monotonic_continuous") in
  let prove_monotonic_continuous = Command.start_proof id_monotonic_continuous goal_kind lemma_monotonic_continuous in
  let proj1 = mkProj1 lemma_monotonic lemma_continuous in
  let proj2 = mkProj2 lemma_monotonic lemma_continuous in
  let f_monotonic mono_cont_proof =
    mkApp (Lazy.force coq_Build_monofun,
     [|ord_typ; ord_typ; f;
       let mono_cont_proof = constr_of_global mono_cont_proof in (proj1 mono_cont_proof)
     |]) in
  let f_monotonic_typ =
    coerce_ord (mkApp (Lazy.force coq_monofunord, [|ord_typ; ord_typ|])) in
  let id_mono_f = id_of_string ("mono_" ^ string_of_f) in
  let declare_monotonic (mono_cont_proof:global_reference) (hook:declaration_hook) =
    declare_constant_hooked id_mono_f f_monotonic_typ (f_monotonic mono_cont_proof) hook
  in
  let f_continuous mono_f mono_cont_proof = mkApp (Lazy.force coq_Build_contfun,
    [|cpo_typ; cpo_typ; constr_of_global mono_f;
      let mono_cont_proof = constr_of_global mono_cont_proof in (mkApp (proj2 mono_cont_proof, [|proj1 mono_cont_proof|]))
    |]) in
  let f_continuous_typ =
    coerce_ord (coerce_cpo (mkApp (Lazy.force coq_contfuncpo, [|cpo_typ; cpo_typ|]))) in
  let id_cont_f = id_of_string ("cont_" ^ string_of_f) in
  let declare_continuous (mono_f:global_reference) (mono_cont_proof:global_reference) (hook:declaration_hook) =
    declare_constant_hooked id_cont_f f_continuous_typ (f_continuous mono_f mono_cont_proof) hook
  in
  let f_lfp (cont_f:global_reference) = let cont_f = constr_of_global cont_f in
    mkApp (Lazy.force coq_f_mono,
     [|
       coerce_cpo (mkApp (Lazy.force coq_contfuncpo, [|cpo_typ; cpo_typ|]));
       coerce_cpo cpo_typ;
       mkApp (Lazy.force coq_contfun_monofun,
         [|
           (mkApp (Lazy.force coq_contfuncpo, [|cpo_typ; cpo_typ|]));
           cpo_typ;
           mkApp (Lazy.force coq_fixp, [|cpo_typ|])
         |]);
       cont_f
     |]) in
  let declare_lfp (cont_f:global_reference) (hook:declaration_hook) =
    declare_constant_hooked id (coerce_ord (coerce_cpo cpo_typ)) (f_lfp cont_f) hook
  in
  let hook_done _ _ = () in
  let hook_lfp _ cont_f = (declare_lfp cont_f hook_done; ()) in
  let hook_declare_continuous mono_cont_proof _ mono_f =
        (declare_continuous mono_f mono_cont_proof hook_lfp; ()) in
  let hook_declare_monotonic _ mono_cont_proof =
        (declare_monotonic mono_cont_proof (hook_declare_continuous mono_cont_proof); ()) in
    prove_monotonic_continuous hook_declare_monotonic;
    Flags.if_verbose (fun () -> msg (Printer.pr_open_subgoals ())) ()
    (* if !pcoq <> None then (Option.get !pcoq).start_proof () *)

(* the main function defining and registering an ffcpo object given its declaration *)
let fcpo_function ((id:identifier), (binders:local_binder list), (typ:constr_expr), (body:constr_expr)) =
  Coqlib.check_required_library ["Coq"; "ffcpo"; "FlatCompletePreorder"];
  let arity = Command.generalize_constr_expr typ binders in
  (* add a function parameter of type [arity] for making recursive calls *)
  let binders = Topconstr.LocalRawAssum([(Util.dummy_loc, Name id)], Default Explicit, arity)::binders in
  let ffcpo_decl = (id, binders, typ, body) in
  let def(*unused*), imps = build_ffcpo ffcpo_decl in
  let _ =
    r := def;
    i := imps;
    let f = declare_functional ffcpo_decl imps in
      (* Pp.msgnl (Printer.pr_constr (constr_of_global f)); *)
      declare_ffcpo id f arity
  in ()

(* let function_fcpo = function_fcpo ffcpo_on_error *)
